'use strict';

const describeIntegrationSupport = require('httplease/test/support/describeIntegrationSupport');
const createCacheFilter = require('../../index').createCacheFilter;

const BODY_CONTENT = {some: 'body'};

describeIntegrationSupport('statusFilter', (support) => {

    let httpSpy;
    let httpClient;

    beforeEach(() => {
        jasmine.clock().install();

        httpSpy = support.httpSpy;
        httpClient = support.httpClient
            .withFilter(createCacheFilter())
            .withPath('/get')
            .withMethodGet();

        jasmine.clock().mockDate(new Date('2016-01-01 10:00:00 UTC'));
    });

    afterEach(() => {
        jasmine.clock().uninstall();
    });

    function setupCacheableResource(headers) {
        headers = Object.assign({
            'Date': '2016-01-01 10:00:00 UTC',
            'Cache-Control': 'max-age=600'
        }, headers);
        httpSpy.getHandler.and.returnValue({
            body: BODY_CONTENT,
            headers
        });
    }

    it('caches responses with max-age', willResolve(() => {
        setupCacheableResource();

        return httpClient.send()
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                jasmine.clock().mockDate(new Date('2016-01-01 10:10:00 UTC'));

                return httpClient.send();
            })
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                expect(httpSpy.getHandler.calls.count()).toBe(1);
            });
    }));

    it('caches responses with expires header', willResolve(() => {
        setupCacheableResource({
            'Cache-Control': undefined,
            'Expires': '2016-01-02 12:00:00 UTC'
        });


        return httpClient.send()
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                jasmine.clock().mockDate(new Date('2016-01-02 11:59:59 UTC'));

                return httpClient.send();
            })
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                expect(httpSpy.getHandler.calls.count()).toBe(1);
            });
    }));

    it('expires cached resources', willResolve(() => {
        setupCacheableResource();

        return httpClient.send()
            .then((response) =>  {
                expect(response.body).toEqual(BODY_CONTENT);

                jasmine.clock().mockDate(new Date('2016-01-01 10:10:01 UTC'));

                return httpClient.send();
            })
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                expect(httpSpy.getHandler.calls.count()).toBe(2);
            });
    }));

    it('respects no-cache even with max-age', willResolve(() => {
        setupCacheableResource({
            'Cache-Control': 'max-age=600;no-cache'
        });

        return httpClient.send()
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                return httpClient.send();
            })
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                expect(httpSpy.getHandler.calls.count()).toBe(2);
            });
    }));

    it('respects no-cache even with expires', willResolve(() => {
        setupCacheableResource({
            'Cache-Control': 'no-cache',
            'Expires': '2099-01-01 00:00:00 UTC'
        });

        return httpClient.send()
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                return httpClient.send();
            })
            .then((response) => {
                expect(response.body).toEqual(BODY_CONTENT);

                expect(httpSpy.getHandler.calls.count()).toBe(2);
            });
    }));

});
